/*----------------------------------------------------------------------
  File    : costab.c
  Contents: read misclassification cost table
  Author  : Christian Borgelt
  History : 2020.09.25 file created
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <time.h>
#ifndef AS_PARSE
#define AS_PARSE
#endif
#ifndef TAB_READ
#define TAB_READ
#endif
#ifndef TAB_WRITE
#define TAB_WRITE
#endif
#include "error.h"
#include "table.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "costab"
#define DESCRIPTION "misclassification cost table reading program"
#define VERSION     "version 0.1 (2020.09.25)        " \
                    "(c) 2020   Christian Borgelt"

/* --- error codes --- */
/* error codes 0 to -5 defined in attset.h */
#define E_OPTION    (-6)        /* unknown option */
#define E_OPTARG    (-7)        /* missing option argument */
#define E_ARGCNT    (-8)        /* wrong number of arguments */
#define E_PARSE     (-9)        /* parse error(s) on domain file */
#define E_CLASS    (-10)        /* class name mismatch */

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char *errmsgs[] = {   /* error messages */
  /* E_NONE     0 */  "no error",
  /* E_NOMEM   -1 */  "not enough memory",
  /* E_FOPEN   -2 */  "cannot open file %s",
  /* E_FREAD   -3 */  "read error on file %s",
  /* E_FWRITE  -4 */  "write error on file %s",
  /* E_STDIN   -5 */  "double assignment of standard input",
  /* E_OPTION  -6 */  "unknown option -%c",
  /* E_OPTARG  -7 */  "missing option argument",
  /* E_ARGCNT  -8 */  "wrong number of arguments",
  /* E_PARSE   -9 */  "parse error(s) on file %s",
  /* E_CLASS  -10 */  "class name mismatch %s : %s",
  /*          -11 */  "unknown error"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static CCHAR    *prgname;       /* program name for error messages */
static SCANNER  *scan   = NULL; /* scanner (for domain file) */
static TABREAD  *tread  = NULL; /* table reader */
static TABWRITE *twrite = NULL; /* table writer */
static ATTSET   *attset = NULL; /* attribute set */
static ATTSET   *astmp  = NULL; /* temporary attribute set */
static TABLE    *costab = NULL; /* misclassification cost table */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/

#ifndef NDEBUG                  /* if debug version */
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (attset) as_delete (attset);    \
  if (astmp)  as_delete (astmp);     \
  if (costab) tab_delete(costab, 0); \
  if (tread)  trd_delete(tread,  1); \
  if (twrite) twr_delete(twrite, 1);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, counters */
  char    *s;                   /* to the traverse options */
  CCHAR   **optarg = NULL;      /* to get option arguments */
  CCHAR   *fn_dom  = NULL;      /* name of domain file */
  CCHAR   *fn_tab  = NULL;      /* name of table  file */
  CCHAR   *recseps = NULL;      /* record  separators */
  CCHAR   *fldseps = NULL;      /* field   separators */
  CCHAR   *blanks  = NULL;      /* blank   characters */
  CCHAR   *comment = NULL;      /* comment characters */
  CCHAR   *nullchs = NULL;      /* null value characters */
  int     mode     = AS_ATT;    /* table file read mode */
  int     *rowmap  = NULL;      /* cost table row    to value id map */
  int     *colmap  = NULL;      /* cost table column to value id map */
  ATT     *att, *src;           /* class attribute(s) */
  TUPLE   *tpl;                 /* to travese the matrix rows */
  CCHAR   *name;                /* name of class attribute */
  ATTID   m, c;                 /* number of attributes/columns */
  TPLID   n, r;                 /* number of data tuples */
  VALID   v, u;                 /* number of attribute values */
  double  w;                    /* weight of data tuples */
  clock_t t;                    /* timer for measurement */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument is given */
    printf("usage: %s [options] tabfile\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-r#      record  separators    (default: \"\\n\")\n");
    printf("-f#      field   separators    (default: \" \\t,\")\n");
    printf("-b#      blank   characters    (default: \" \\t\\r\")\n");
    printf("-C#      comment characters    (default: \"#\")\n");
    printf("hdrfile  file containing table header "
                    "(field names)\n");
    printf("tabfile  table file to read "
                    "(field names in first record)\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (1) {               /* traverse characters */
        switch (*s++) {         /* evaluate option */
          case 'r': optarg = &recseps;     break;
          case 'f': optarg = &fldseps;     break;
          case 'b': optarg = &blanks;      break;
          case 'C': optarg = &comment;     break;
          default : error(E_OPTION, *--s); break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fn_dom = s;      break;
        case  1: fn_tab = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note filename */
    }
  }
  if (optarg) error(E_OPTARG);  /* check option argument */
  if (k != 2) error(E_ARGCNT);  /* check number of arguments */
  fputc('\n', stderr);          /* terminate the startup message */

  /* --- parse domain descriptions --- */
  attset = as_create("domains", att_delete);
  if (!attset)  error(E_NOMEM); /* create an attribute set */
  scan = scn_create();          /* create a scanner */
  if (!scan) error(E_NOMEM);    /* for the domain file */
  t = clock();                  /* start timer, open intput file */
  if (scn_open(scan, NULL, fn_dom) != 0)
    error(E_FOPEN, scn_name(scan));
  fprintf(stderr, "reading %s ... ", scn_name(scan));
  if ((as_parse(attset, scan, AT_ALL, 0) != 0)
  ||  !scn_eof(scan, 1))        /* parse domain descriptions */
    error(E_PARSE, scn_name(scan));
  scn_delete(scan, 1);          /* delete the scanner and */
  scan = NULL;                  /* print a success message */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", as_attcnt(attset));
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- read cost table --- */
  tread = trd_create();         /* create a table reader and */
  if (!tread)  error(E_NOMEM);  /* configure the characters */
  trd_allchs(tread, recseps, fldseps, blanks, "", comment);
  astmp = as_create("domains", att_delete);
  if (!astmp) error(E_NOMEM);   /* create an attribute set */
  costab = tab_create("costab", astmp, tpl_delete);
  if (!costab) error(E_NOMEM);  /* create a data table */
  t = clock();                  /* start timer, open input file */
  if (trd_open(tread, NULL, fn_tab) != 0)
    error(E_FOPEN, trd_name(tread));
  fprintf(stderr, "reading %s ... ", trd_name(tread));
  k = tab_read(costab, tread, mode);
  if (k < 0) error(-k, tab_errmsg(costab, NULL, 0));
  trd_delete(tread, 1);         /* read the table body and */
  tread = NULL;                 /* delete the table reader */
  m = tab_attcnt(costab);       /* get the number of attributes */
  n = tab_tplcnt(costab);       /* and the number of data tuples */
  w = tab_tplwgt(costab);       /* and their total weight */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s),", m);
  fprintf(stderr, " %"TPLID_FMT, n);
  if (w != (double)n) fprintf(stderr, "/%g", w);
  fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));

  /* --- conver column tapes --- */
  for (i = tab_colcnt(costab); --i > 0; )
    tab_colconv(costab, i, AT_FLT);

  /* --- show cost table --- */
  twrite = twr_create();        /* create a table writer and */
  if (!twrite) error(E_NOMEM);  /* configure the characters */
  twr_xchars(twrite, recseps, fldseps, blanks, nullchs);
  twr_open(twrite, stdout, NULL);
  k = tab_write(costab, twrite, AS_ATT|AS_ALIGN);
  if (k < 0) error(k, twr_name(twrite));
  twr_delete(twrite, 0);        /* write the data table */
  twrite = NULL;                /* to standard output and */

  /* --- get row and column maps --- */
  att  = as_att(attset, as_attcnt(attset)-1);
  name = att_name(tab_col(costab, 0));
  if (strcmp(att_name(att), name) != 0)
    error(E_CLASS, att_name(att), name);
  v = att_valcnt(att);          /* check class name, get values */
  rowmap = (int*)malloc((size_t)(v+v+2) *sizeof(int));
  if (!rowmap) error(E_NOMEM);  /* create row and column maps */
  colmap = rowmap +v+1;         /* and organize the memory */
  for (u = v+1; --u >= 0; )     /* initialize maps to -1 */
    rowmap[u] = colmap[u] = -1; /* (no cost row/column given) */
  src = tab_col(costab, 0);     /* get class attribute in cost table */
  for (r = tab_rowcnt(costab); --r >= 0; ) {
    tpl  = tab_tpl(costab, r);  /* traverse the cost table rows */
    name = att_valname(src, tpl_colval(tpl, 0)->n);
    if (strcmp(name, "?") == 0) { rowmap[v] = r; continue; }
    if ((u = att_valid(att, name)) >= 0) rowmap[u] = r;
  }                             /* collect row ids for classes */
  for (c = tab_colcnt(costab); --c > 0; ) {
    name = tab_colname(costab, c);
    if (strcmp(name, "?") == 0) { colmap[v] = c; continue; }
    if ((u = att_valid(att, name)) >= 0) colmap[u] = c;
  }                             /* collect column ids for classes */
  for (u = 0; u < v; u++) {     /* check for unset clumns */
    if (rowmap[u] < 0) rowmap[u] = rowmap[v];
    if (colmap[u] < 0) colmap[u] = colmap[v];
  }                             /* map to additional row/column */

  for (i = 0; i <= v; i++) {
    name = (i < v) ? att_valname(att, i) : "?";
    printf("%-16s: %d %d\n", name, rowmap[i], colmap[i]);
  }

  free(rowmap);

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */
